def generate_unique_filename(output_path):
    base = output_path.stem
    suffix = output_path.suffix
    directory = output_path.parent

    counter = 1
    new_path = output_path
    while new_path.exists():
        new_path = directory / f"{base}_{counter}{suffix}"
        counter += 1

    return new_path



import sys
import os
from elevenlabs.client import ElevenLabs
from elevenlabs import save

ELEVEN_TEXT_LIMIT_NONSIGNED=500

def process_article_elevenlabs(text, filename, model, voice):
    try:
        api_key = os.environ['ELEVEN_API_KEY']
        client = ElevenLabs(api_key=api_key)
    except:
        print("Will try to use EleveLabs without API key.")

        if len(text) > ELEVEN_TEXT_LIMIT_NONSIGNED:
            print(f"""
This request's text has {len(text)} characters and exceeds the character limit
of 500 characters for non signed in accounts.
"""
            )
            sys.exit(0)
        else:
            client = ElevenLabs()

    audio = client.generate(
      text=text,
      voice=voice,
      model=model
    )

    save(audio, filename)



from click.testing import CliRunner
from audio_reads.cli import cli
from audio_reads.chunks import TEXT_SEND_LIMIT, split_text
from audio_reads.article import get_article_content
from pathlib import Path
import pytest

ARTICLE_URL_HTML = "https://blog.kubetools.io/kopylot-an-ai-powered-kubernetes-assistant-for-devops-developers/"
ARTICLE_URL_JS = (
    "https://lab.scub.net/architecture-patterns-the-circuit-breaker-8f79280771f1"
)
ARTICLES_FILE_PATH = "/tmp/articles-file-list.txt"


@pytest.fixture
def setup_article_file():
    with open(ARTICLES_FILE_PATH, "w") as article_file_list:
        for _ in range(2):
            article_file_list.write(ARTICLE_URL_HTML + "\n")
    yield ARTICLES_FILE_PATH
    Path(ARTICLES_FILE_PATH).unlink()  # Clean up the file after the test


def test_split_text():
    text = "This is a test text. " * 300  # Creating a long text to ensure it gets split
    chunks = split_text(text)

    # Ensure that the text is split into more than one chunk
    assert len(chunks) > 1

    # Ensure that each chunk is within the limit
    for chunk in chunks:
        assert len(chunk) <= TEXT_SEND_LIMIT


def test_get_article_content():
    text, title = get_article_content(ARTICLE_URL_HTML)

    # Check if a known phrase is in the text and title
    assert (
        "KoPylot\xa0is a cloud-native application performance monitoring (APM) solution that runs on Kubernetes"
        in text
    )
    assert "KoPylot" in title  # Checking a part of the title to ensure it's correct


@pytest.mark.parametrize(
    "url, expected_exit_code",
    [
        (ARTICLE_URL_HTML, 0),
        (ARTICLE_URL_JS, 0),
    ],
)
def test_process_article_openai(url, expected_exit_code):
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--url",
            url,
            "--directory",
            "/tmp",
            "--audio-format",
            "mp3",
            "--model",
            "tts-1",
            "--voice",
            "alloy",
            "--strip",
            "5",  # Strip the text by # of chars to reduce costs during testing
        ],
        catch_exceptions=False,  # Allow exceptions to propagate
    )

    assert result.exit_code == expected_exit_code

    output_audio_path = next(
        Path("/tmp").glob("*.mp3")
    )  # Find the generated audio file
    assert output_audio_path.exists()

    # Clean up
    output_audio_path.unlink()


def test_process_article_elevenlabs():
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--url",
            ARTICLE_URL_HTML,
            "--vendor",
            "elevenlabs",
            "--directory",
            "/tmp",
            "--strip",
            "5",  # Strip the text by # of chars to reduce costs during testing
        ],
        catch_exceptions=False,  # Allow exceptions to propagate
    )

    assert result.exit_code == 0

    output_audio_path = next(
        Path("/tmp").glob("*.mp3")
    )  # Find the generated audio file
    assert output_audio_path.exists()

    # Clean up
    output_audio_path.unlink()


def test_process_article_openai_file_list(setup_article_file):
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--file-url-list",
            setup_article_file,
            "--directory",
            "/tmp",
            "--audio-format",
            "mp3",
            "--model",
            "tts-1",
            "--voice",
            "alloy",
            "--strip",
            "5",  # Strip the text by # of chars to reduce costs during testing
        ],
        catch_exceptions=False,  # Allow exceptions to propagate
    )

    assert result.exit_code == 0

    # Find the generated audio files
    output_audio_paths = list(Path("/tmp").glob("*.mp3"))
    assert len(output_audio_paths) == 2  # Ensure two audio files are created

    for output_audio_path in output_audio_paths:
        assert output_audio_path.exists()
        # Clean up
        output_audio_path.unlink()



# * OpenIA limit in chars for one request is 4096.
# * For ElevenLabs it's 2500 for non subscribed users and 5000 for
# subscribed ones.
#
# Thus, let's set common limit to 2500
TEXT_SEND_LIMIT = 2500


def split_text(text, limit=TEXT_SEND_LIMIT):
    words = text.split()
    chunks = []
    current_chunk = words[0]

    for word in words[1:]:
        if len(current_chunk) + len(word) + 1 <= limit:
            current_chunk += " " + word
        else:
            chunks.append(current_chunk)
            current_chunk = word
    chunks.append(current_chunk)

    return chunks



from bs4 import BeautifulSoup
from readability import Document
from playwright.sync_api import sync_playwright
import requests
from .common import RenderError


def is_js_required(soup):
    """Check if the content indicates JS is required"""
    # Check for elements that are often loaded by JavaScript
    js_indicators = [
        ("div", {"id": "app"}),
        ("div", {"id": "root"}),
        ("script", {}),
        ("noscript", {}),
    ]

    for tag_name, attr_dict in js_indicators:
        tag_elements = soup.find_all(tag_name, attr_dict)
        if tag_elements:
            return True

    return False


def fetch_content_with_requests(url):
    try:
        response = requests.get(url, timeout=10)
        response.raise_for_status()

        soup = BeautifulSoup(
            response.text, "html.parser"
        )  # <-- Create soup from response.text
        doc = Document(soup)  # <-- Pass the soup object to Document

        text = soup.get_text()
        title = doc.title()

        print(f"soup type is {type(soup)}")
        print(f"soup value is {soup}")
        if is_js_required(soup):
            raise RenderError("JavaScript is required to load this page.")

        return text, title
    except requests.RequestException as e:
        raise RenderError(f"Request failed: {e}")
    except Exception as e:
        raise RenderError(f"Parsing failed: {e}")


def fetch_content_with_playwright(url):
    with sync_playwright() as p:
        browser = p.chromium.launch(headless=True)
        page = browser.new_page()

        # Navigate to the URL
        page.goto(url)

        # Wait for the page to load completely
        page.wait_for_load_state("networkidle")

        # Get the rendered HTML
        html = page.content()

        # Use BeautifulSoup and readability to extract text and title
        doc = Document(html)
        soup = BeautifulSoup(doc.summary(), "html.parser")
        title = doc.title()
        text = soup.get_text()

        browser.close()

        return text, title


def get_article_content(url):
    try:
        print("Attempting to fetch and parse using requests.")
        return fetch_content_with_requests(url)
    except RenderError:
        print("Using Playwright to render the page.")
        return fetch_content_with_playwright(url)



from pathlib import Path
import click
import os
import re

from .article import get_article_content
from .common import RenderError
from .elevenlabs import process_article_elevenlabs
from .openai import process_article_openai


def format_filename(title, format):
    # Replace special characters with dashes and convert to lowercase
    formatted_title = re.sub(r"\W+", "-", title).strip("-").lower()
    return f"{formatted_title}.{format}"


# Define models depending on the AI vendor
def validate_models(ctx, param, value):
    if value is None:
        return value

    try:
        vendor = ctx.params["vendor"]
    except:
        vendor = "openai"

    if vendor == "elevenlabs":
        choices = ["eleven_monolingual_v1"]
    else:
        choices = ["tts-1", "tts-1-hd"]

    if value not in choices:
        raise click.BadParameter(f"Invalid choice: {value}. Allowed choices: {choices}")
    return value


def validate_voice(ctx, param, value):
    if value is None:
        return value

    try:
        vendor = ctx.params["vendor"]
    except:
        vendor = "openai"

    if vendor == "elevenlabs":
        choices = ["Nicole"]
    else:
        choices = ["alloy", "echo", "fable", "onyx", "nova", "shimmer"]

    if value not in choices:
        raise click.BadParameter(f"Invalid choice: {value}. Allowed choices: {choices}")
    return value


@click.command()
@click.option("--url", type=str, help="URL of the article to be fetched.")
@click.option(
    "--vendor",
    type=click.Choice(["openai", "elevenlabs"]),
    default="openai",
    help="Choose vendor to use to convert text to audio.",
)
@click.option(
    "--file-url-list",
    type=click.Path(exists=True, dir_okay=False, readable=True),
    help="Path to a file with URLs placed on every new line.",
)
@click.option(
    "--directory",
    type=click.Path(exists=False, file_okay=False, writable=True),
    default=".",
    help="Directory where the output audio file will be saved. The filename will be derived from the article title.",
)
@click.option(
    "--model",
    callback=validate_models,
    default=None,
    help="The model to be used for text-to-speech conversion.",
)
@click.option(
    "--voice",
    callback=validate_voice,
    default=None,
    help="""
    OpenIA voices: alloy, echo, fable, onyx, nova, shimmer;
    ElevenLabs voices: Nicole.
    """,
)
@click.option(
    "--strip",
    type=click.IntRange(5, 2000),
    help="By what number of chars to strip the text to send to OpenAI.",
)
@click.option(
    "--audio-format",
    type=click.Choice(["mp3", "opus", "aac", "flac", "pcm"]),
    default="mp3",
    help="The audio format for the output file. Default is mp3.",
)
def cli(vendor, url, file_url_list, directory, audio_format, model, voice, strip):
    if not url and not file_url_list:
        raise click.UsageError("You must provide either --url or --file-url-list.")

    # Set model and voice based on the API vendor
    if vendor == "elevenlabs":
        model = model or "eleven_monolingual_v1"
        voice = voice or "Nicole"
    elif vendor == "openai":
        model = model or "tts-1"
        voice = voice or "alloy"

    urls = []
    if url:
        urls.append(url)
    if file_url_list:
        with open(file_url_list, "r") as f:
            urls.extend([line.strip() for line in f if line.strip()])

    for url in urls:
        text, title = get_article_content(url)

        # Strip text by number of chars set
        if strip:
            text = text[:strip]

        # Create directory if it does not exist
        os.makedirs(directory, exist_ok=True)
        print(f"Processing article with `{title}` to audio..")
        filename = Path(directory) / f"{format_filename(title, audio_format)}"

        if vendor == "openai":
            process_article_openai(text, filename, model, voice)
        elif vendor == "elevenlabs":
            process_article_elevenlabs(text, filename, model, voice)


if __name__ == "__main__":
    cli()
